# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import abc
import pandas as pd
import torch
import torch.nn as nn
import torch.distributed as dist
import torch.nn.functional as F
from dataclasses import dataclass, field, fields
from itertools import chain
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


@dataclass
class TensorsMixin:
    def __post_init__(
        self,
    ):
        if not hasattr(self, "__tensors__"):
            self.__tensors__ = {}
        for key, tensor in self.__tensors__.items():
            assert isinstance(tensor, torch.Tensor)
            setattr(self, key, tensor)

    @classmethod
    def union(cls, *list_tensors, dim=0):
        if len(list_tensors) == 0:
            return cls()
        first_tensors = list_tensors[0]
        keys = list(first_tensors.__tensors__.keys())
        for tensors in list_tensors:
            assert keys == list(tensors.__tensors__.keys())

        new_tensors = {}
        for key in keys:
            new_tensor = [tensors.__tensors__[key] for tensors in list_tensors]
            if new_tensor[0] is not None:
                new_tensors[key] = torch.cat(new_tensor, dim=dim)
            else:
                new_tensors[key] = None

        return cls(**new_tensors)

    @classmethod
    def stack(cls, *list_tensors, dim=0):
        if len(list_tensors) == 0:
            return cls()
        first_tensors = list_tensors[0]
        keys = list(first_tensors.__tensors__.keys())
        for tensors in list_tensors:
            assert keys == list(tensors.__tensors__.keys())

        new_tensors = {}
        for key in keys:
            new_tensor = [tensors.__tensors__[key] for tensors in list_tensors]
            if new_tensor[0] is not None:
                new_tensors[key] = torch.stack(new_tensor, dim=dim)
            else:
                new_tensors[key] = None

        return cls(**new_tensors)

    def add(self, tensors: Union[Dict, "TensorsMixin"]):
        if isinstance(tensors, dict):
            tensors_dict = tensors
        else:
            tensors_dict = tensors.__tensors__
        for key, value in tensors_dict.items():
            assert isinstance(value, torch.Tensor)
            if key in self.__tensors__:
                assert ValueError(f"{key} already in the tensor.")
            else:
                self.__tensors__[key] = value
                setattr(self, key, value)

    def cpu(self, inplace=False):
        new_tensors = {}
        for key, value in self.__tensors__.items():
            new_value = value.cpu() if value is not None else None
            new_tensors[key] = new_value
        if inplace:
            for key, value in new_tensors.items():
                self.__tensors__[key] = value
            return
        return type(self)(**new_tensors)

    def cuda(self, inplace=False):
        new_tensors = {}
        for key, value in self.__tensors__.items():
            new_value = value.cuda() if value is not None else None
            new_tensors[key] = new_value
        if inplace:
            for key, value in new_tensors.items():
                self.__tensors__[key] = value
            return
        return type(self)(**new_tensors)

    def sync(self, dim=0, inplace=False):
        new_tensors = {}
        for key, value in self.__tensors__.items():
            if value is not None:
                new_value = [value.clone() for _ in range(dist.get_world_size())]
                dist.all_gather(new_value, value)
                new_value = torch.cat(new_value, dim=dim)
            else:
                new_value = None
            new_tensors[key] = new_value
        if inplace:
            for key, value in new_tensors.items():
                self.__tensors__[key] = value
            return
        return type(self)(**new_tensors)

    def dict(self):
        return self.__tensors__


@dataclass
class ListTensorsMixin:
    def __post_init__(
        self,
    ):
        if not hasattr(self, "__list_tensors__"):
            self.__list_tensors__ = {}
        new_list_tensors = {}
        for key, value in self.__list_tensors__.items():
            if isinstance(value, torch.Tensor):
                value = [value]
            new_list_tensors[key] = value
        self.__list_tensors__ = new_list_tensors
        for key, list_tensor in self.__list_tensors__.items():
            assert isinstance(list_tensor, list)
            assert all([isinstance(tensor, torch.Tensor) for tensor in list_tensor])
            setattr(self, key, list_tensor)

    @classmethod
    def union(cls, *list_of_list_tensors, dim=0):
        if len(list_of_list_tensors) == 0:
            return cls()
        first_list_tensors = list_of_list_tensors[0]
        keys = list(first_list_tensors.__list_tensors__.keys())
        for list_tensors in list_of_list_tensors:
            assert keys == list(list_tensors.__list_tensors__.keys())

        new_list_tensors = {}
        for key in keys:
            new_list_tensor = [
                list_tensors.__list_tensors__[key]
                for list_tensors in list_of_list_tensors
            ]
            if new_list_tensor[0] is not None:
                new_list_tensors[key] = list(chain.from_iterable(new_list_tensor))
            else:
                new_list_tensors[key] = None

        return cls(**new_list_tensors)

    @classmethod
    def stack(cls, *list_of_list_tensors, dim=0):
        if len(list_of_list_tensors) == 0:
            return cls()
        first_list_tensors = list_of_list_tensors[0]
        keys = list(first_list_tensors.__list_tensors__.keys())
        for list_tensors in list_of_list_tensors:
            assert keys == list(list_tensors.__list_tensors__.keys())

        new_list_tensors = {}
        for key in keys:
            new_list_tensor = [
                list_tensors.__list_tensors__[key]
                for list_tensors in list_of_list_tensors
            ]
            if new_list_tensor[0] is not None:
                assert all([len(list_tensor) == 1 for list_tensor in new_list_tensor])
                new_list_tensors[key] = list(chain.from_iterable(new_list_tensor))
            else:
                new_list_tensors[key] = None

        return cls(**new_list_tensors)

    def add(self, list_tensors: Union[Dict, "ListTensorsMixin"]):
        if isinstance(list_tensors, dict):
            list_tensors_dict = list_tensors
        else:
            list_tensors_dict = list_tensors.__list_tensors__
        for key, value in list_tensors_dict.items():
            assert isinstance(value, list)
            if key in self.__list_tensors__:
                assert ValueError(f"{key} already in the list tensor.")
            else:
                if isinstance(value, torch.Tensor):
                    value = [value]
                self.__list_tensors__[key] = value
                setattr(self, key, value)

    def cpu(self, inplace=False):
        new_list_tensors = {}
        for key, value in self.__list_tensors__.items():
            if value is not None and value[0] is not None:
                new_value = [_value.cpu() for _value in value]
                new_list_tensors[key] = new_value
            else:
                new_list_tensors[key] = None
        if inplace:
            for key, value in new_list_tensors.items():
                self.__tensors__[key] = value
            return
        return type(self)(**new_list_tensors)

    def cuda(self, inplace=False):
        new_list_tensors = {}
        for key, value in self.__list_tensors__.items():
            if value is not None and value[0] is not None:
                new_value = [_value.cuda() for _value in value]
                new_list_tensors[key] = new_value
            else:
                new_list_tensors[key] = None
        if inplace:
            for key, value in new_list_tensors.items():
                self.__list_tensors__[key] = value
            return
        return type(self)(**new_list_tensors)

    def sync(self, inplace=False):
        new_list_tensors = {}
        for key, value in self.__list_tensors__.items():
            if value is not None:
                new_value = [[] for _ in range(dist.get_world_size())]
                dist.all_gather_object(new_value, value)
                new_value = list(chain.from_iterable(new_value))
            else:
                new_value = None
            new_list_tensors[key] = new_value
        if inplace:
            for key, value in new_list_tensors.items():
                self.__list_tensors__[key] = value
            return
        return type(self)(**new_list_tensors)

    def dict(self):
        return self.__list_tensors__


class CombineTensorsMixin(TensorsMixin, ListTensorsMixin):
    def __init__(
        self,
        dict_of_tensors: Dict[str, torch.Tensor] = dict(),
        dict_of_list_tensors: Dict[str, List[torch.Tensor]] = dict(),
    ):
        self.__tensors__ = dict_of_tensors
        self.__list_tensors__ = dict_of_list_tensors
        self.__post_init__()

    def __post_init__(
        self,
    ):
        TensorsMixin.__post_init__(self)
        ListTensorsMixin.__post_init__(self)

    @classmethod
    def union(cls, *list_of_mix_tensors, dim=0):
        if len(list_of_mix_tensors) == 0:
            return cls()
        first_mix_tensors = list_of_mix_tensors[0]
        tensors_keys = list(first_mix_tensors.__tensors__.keys())
        list_tensors_keys = list(first_mix_tensors.__list_tensors__.keys())
        for mix_tensors in list_of_mix_tensors:
            assert tensors_keys == list(mix_tensors.__tensors__.keys())
            assert list_tensors_keys == list(mix_tensors.__list_tensors__.keys())

        new_tensors = {}
        for key in tensors_keys:
            new_tensor = [
                mix_tensors.__tensors__[key] for mix_tensors in list_of_mix_tensors
            ]
            if new_tensor[0] is not None:
                new_tensors[key] = torch.cat(new_tensor, dim=dim)
            else:
                new_tensors[key] = None

        new_list_tensors = {}
        for key in list_tensors_keys:
            new_list_tensor = [
                mix_tensors.__list_tensors__[key] for mix_tensors in list_of_mix_tensors
            ]
            if new_list_tensor[0] is not None:
                new_list_tensors[key] = list(chain.from_iterable(new_list_tensor))
            else:
                new_list_tensors[key] = None

        return cls(dict_of_tensors=new_tensors, dict_of_list_tensors=new_list_tensors)

    @classmethod
    def stack(cls, *list_of_mix_tensors, dim=0):
        if len(list_of_mix_tensors) == 0:
            return cls()
        first_mix_tensors = list_of_mix_tensors[0]
        tensors_keys = list(first_mix_tensors.__tensors__.keys())
        list_tensors_keys = list(first_mix_tensors.__list_tensors__.keys())
        for mix_tensors in list_of_mix_tensors:
            assert tensors_keys == list(mix_tensors.__tensors__.keys())
            assert list_tensors_keys == list(mix_tensors.__list_tensors__.keys())

        new_tensors = {}
        for key in tensors_keys:
            new_tensor = [
                mix_tensors.__tensors__[key] for mix_tensors in list_of_mix_tensors
            ]
            if new_tensor[0] is not None:
                new_tensors[key] = torch.stack(new_tensor, dim=dim)
            else:
                new_tensors[key] = None

        new_list_tensors = {}
        for key in list_tensors_keys:
            new_list_tensor = [
                mix_tensors.__list_tensors__[key] for mix_tensors in list_of_mix_tensors
            ]
            if new_list_tensor[0] is not None:
                new_list_tensors[key] = new_list_tensor
            else:
                new_list_tensors[key] = None

        return cls(dict_of_tensors=new_tensors, dict_of_list_tensors=new_list_tensors)

    def add(self, mix_tensors: Union[TensorsMixin, ListTensorsMixin]):
        if isinstance(mix_tensors, TensorsMixin):
            TensorsMixin.add(self, mix_tensors.__tensors__)
        if isinstance(mix_tensors, ListTensorsMixin):
            ListTensorsMixin.add(self, mix_tensors.__list_tensors__)

    def cpu(self, inplace=False):
        new_tensors = TensorsMixin.cpu(self, inplace=inplace)
        new_list_tensors = ListTensorsMixin.cpu(self, inplace=inplace)
        if inplace:
            return
        return type(self)(
            dict_of_tensors=new_tensors.__tensors__,
            dict_of_list_tensors=new_list_tensors.__list_tensors__,
        )

    def cuda(self, inplace=False):
        new_tensors = TensorsMixin.cuda(self, inplace=inplace)
        new_list_tensors = ListTensorsMixin.cuda(self, inplace=inplace)
        if inplace:
            return
        return type(self)(
            dict_of_tensors=new_tensors.__tensors__,
            dict_of_list_tensors=new_list_tensors.__list_tensors__,
        )

    def sync(self, inplace=False):
        new_tensors = TensorsMixin.sync(self, inplace=inplace)
        new_list_tensors = ListTensorsMixin.sync(self, inplace=inplace)
        if inplace:
            return
        return type(self)(
            dict_of_tensors=new_tensors.__tensors__,
            dict_of_list_tensors=new_list_tensors.__list_tensors__,
        )

    def dict(self):
        return {**self.__tensors__, **self.__list_tensors__}


class ModelInputs(metaclass=abc.ABCMeta):
    pass


class ModelOutputs(metaclass=abc.ABCMeta):
    pass


class ModelTargets(metaclass=abc.ABCMeta):
    pass


@dataclass(init=False)
class TensorsInputs(ModelInputs, TensorsMixin):
    def __init__(self, inputs: Optional[Dict] = dict(), **kwargs):
        self.__tensors__ = {}
        for k, v in {**inputs, **kwargs}.items():
            if not hasattr(self, k):
                setattr(self, k, v)
            self.__tensors__[k] = v


class TensorsOutputs(ModelOutputs, TensorsMixin):
    def __post_init__(self):
        self.__tensors__ = {}
        for f in fields(self):
            self.__tensors__[f.name] = getattr(self, f.name)
        super().__post_init__()


class TensorsTargets(ModelTargets, TensorsMixin):
    def __post_init__(self):
        self.__tensors__ = {}
        for f in fields(self):
            self.__tensors__[f.name] = getattr(self, f.name)
        super().__post_init__()


@dataclass(init=False)
class ListTensorsInputs(ModelInputs, ListTensorsMixin):
    def __init__(self, inputs: Optional[Dict] = dict(), **kwargs):
        self.__list_tensors__ = {}
        for k, v in {**inputs, **kwargs}.items():
            if not hasattr(self, k):
                setattr(self, k, v)
            self.__list_tensors__[k] = v
        super().__post_init__()

    def __post_init__(self):
        for f in fields(self):
            self.__list_tensors__[f.name] = getattr(self, f.name)
        super().__post_init__()


class ListTensorsOutputs(ModelOutputs, ListTensorsMixin):
    def __post_init__(self):
        self.__list_tensors__ = {}
        for f in fields(self):
            self.__list_tensors__[f.name] = getattr(self, f.name)
        super().__post_init__()


class ListTensorsTargets(ModelTargets, ListTensorsMixin):
    def __post_init__(self):
        self.__list_tensors__ = {}
        for f in fields(self):
            self.__list_tensors__[f.name] = getattr(self, f.name)
        super().__post_init__()


class CombineTensorsInputs(ModelInputs, CombineTensorsMixin):
    def __init__(
        self,
        dict_of_tensors: Dict[str, torch.Tensor] = dict(),
        dict_of_list_tensors: Dict[str, List[torch.Tensor]] = dict(),
    ):
        CombineTensorsMixin.__init__(
            self,
            dict_of_tensors=dict_of_tensors,
            dict_of_list_tensors=dict_of_list_tensors,
        )


class CombineTensorsOutputs(ModelOutputs, CombineTensorsMixin):
    def __init__(
        self,
        dict_of_tensors: Dict[str, torch.Tensor] = dict(),
        dict_of_list_tensors: Dict[str, List[torch.Tensor]] = dict(),
    ):
        CombineTensorsMixin.__init__(
            self,
            dict_of_tensors=dict_of_tensors,
            dict_of_list_tensors=dict_of_list_tensors,
        )


class CombineTensorsTargets(ModelTargets, CombineTensorsMixin):
    def __init__(
        self,
        dict_of_tensors: Dict[str, torch.Tensor] = dict(),
        dict_of_list_tensors: Dict[str, List[torch.Tensor]] = dict(),
    ):
        CombineTensorsMixin.__init__(
            self,
            dict_of_tensors=dict_of_tensors,
            dict_of_list_tensors=dict_of_list_tensors,
        )


@dataclass
class LossOutputs(TensorsOutputs):
    loss: torch.Tensor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import re
import sys
import platform
import logging
import torch
from itertools import chain
from setuptools import find_packages, setup

enabled_extensions = os.environ.get("UNITORCH_EXTENSIONS", None)

if enabled_extensions is not None:
    enabled_extensions = re.split(";|,| ", enabled_extensions)
    enabled_extensions = [ext.upper() for ext in enabled_extensions]
else:
    enabled_extensions = []

if torch.cuda.is_available():
    from torch.utils.cpp_extension import BuildExtension, CUDAExtension

    ngram_cuda_extension = CUDAExtension(
        "unitorch.clib.ngram_repeat_block_cuda",
        [
            "unitorch/clib/cuda/ngram_repeat_block/module.cpp",
            "unitorch/clib/cuda/ngram_repeat_block/kernel.cu",
        ],
    )

    all_extensions = {
        "NGRAM": ngram_cuda_extension,
    }

else:
    all_extensions = {}


setup(
    ext_modules=list(
        chain(
            [all_extensions[ext] for ext in enabled_extensions if ext in all_extensions]
        )
    ),
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import diffusers.schedulers as schedulers
from transformers import CLIPTextConfig, CLIPTextModel, CLIPTextModelWithProjection
from diffusers.schedulers import SchedulerMixin
from diffusers.models.attention_processor import (
    AttnAddedKVProcessor,
    AttnAddedKVProcessor2_0,
    LoRAAttnAddedKVProcessor,
    LoRAAttnProcessor,
    LoRAAttnProcessor2_0,
    SlicedAttnAddedKVProcessor,
)
from diffusers.models import (
    ControlNetModel,
    UNet2DModel,
    UNet2DConditionModel,
    AutoencoderKL,
)
from diffusers.pipelines import (
    StableDiffusionXLControlNetPipeline,
    StableDiffusionXLControlNetImg2ImgPipeline,
    StableDiffusionXLControlNetInpaintPipeline,
)
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)


class GenericControlNetXLModel(GenericModel, QuantizationMixin):
    prefix_keys_in_state_dict = {
        # unet weights
        "^add_embedding.*": "unet.",
        "^conv_in.*": "unet.",
        "^conv_norm_out.*": "unet.",
        "^conv_out.*": "unet.",
        "^time_embedding.*": "unet.",
        "^up_blocks.*": "unet.",
        "^mid_block.*": "unet.",
        "^down_blocks.*": "unet.",
        # vae weights
        "^encoder.*": "vae.",
        "^decoder.*": "vae.",
        "^post_quant_conv.*": "vae.",
        "^quant_conv.*": "vae.",
    }

    replace_keys_in_state_dict = {
        "\.query\.": ".to_q.",
        "\.key\.": ".to_k.",
        "\.value\.": ".to_v.",
        "\.proj_attn\.": ".to_out.0.",
    }

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_unet_encoder: Optional[bool] = True,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        super().__init__()
        self.seed = seed
        self.num_train_timesteps = num_train_timesteps
        self.num_infer_timesteps = num_infer_timesteps

        config_dict = json.load(open(config_path))
        if image_size is not None:
            config_dict.update({"sample_size": image_size})
        if in_channels is not None:
            config_dict.update({"in_channels": in_channels})
        if out_channels is not None:
            config_dict.update({"out_channels": out_channels})
        self.unet = UNet2DConditionModel.from_config(config_dict)
        text_config = CLIPTextConfig.from_json_file(text_config_path)
        self.text = CLIPTextModel(text_config)
        text_config2 = CLIPTextConfig.from_json_file(text2_config_path)
        self.text2 = CLIPTextModelWithProjection(text_config2)
        vae_config_dict = json.load(open(vae_config_path))
        self.vae = AutoencoderKL.from_config(vae_config_dict)
        controlnet_config_dict = json.load(open(controlnet_config_path))
        self.controlnet = ControlNetModel.from_config(controlnet_config_dict)

        scheduler_config_dict = json.load(open(scheduler_config_path))
        scheduler_class_name = scheduler_config_dict.get("_class_name", "DDPMScheduler")
        assert hasattr(schedulers, scheduler_class_name)
        scheduler_class = getattr(schedulers, scheduler_class_name)
        assert issubclass(scheduler_class, SchedulerMixin)
        scheduler_config_dict["num_train_timesteps"] = num_train_timesteps
        self.scheduler = scheduler_class.from_config(scheduler_config_dict)

        if freeze_vae_encoder:
            for param in self.vae.parameters():
                param.requires_grad = False

        if freeze_text_encoder:
            for param in self.text.parameters():
                param.requires_grad = False

            for param in self.text2.parameters():
                param.requires_grad = False

        if freeze_unet_encoder:
            for param in self.unet.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(
                self.quant_config,
                ignore_modules=["lm_head", "unet", "vae", "controlnet"],
            )

        if lora_r is not None:
            for param in self.unet.parameters():
                param.requires_grad = False
            self.enable_lora(lora_r=lora_r)

        self.scheduler.set_timesteps(num_inference_steps=self.num_infer_timesteps)

    def enable_lora(self, lora_r: Optional[int] = 4):
        lora_attn_procs = {}
        for name, attn_processor in self.unet.attn_processors.items():
            cross_attention_dim = (
                None
                if name.endswith("attn1.processor")
                else self.unet.config.cross_attention_dim
            )
            if name.startswith("mid_block"):
                hidden_size = self.unet.config.block_out_channels[-1]
            elif name.startswith("up_blocks"):
                block_id = int(name[len("up_blocks.")])
                hidden_size = list(reversed(self.unet.config.block_out_channels))[
                    block_id
                ]
            elif name.startswith("down_blocks"):
                block_id = int(name[len("down_blocks.")])
                hidden_size = self.unet.config.block_out_channels[block_id]

            if isinstance(
                attn_processor,
                (
                    AttnAddedKVProcessor,
                    SlicedAttnAddedKVProcessor,
                    AttnAddedKVProcessor2_0,
                ),
            ):
                lora_attn_processor_class = LoRAAttnAddedKVProcessor
            else:
                lora_attn_processor_class = (
                    LoRAAttnProcessor2_0
                    if hasattr(F, "scaled_dot_product_attention")
                    else LoRAAttnProcessor
                )

            module = lora_attn_processor_class(
                hidden_size=hidden_size,
                cross_attention_dim=cross_attention_dim,
                rank=lora_r,
            )

            lora_attn_procs[name] = module

        self.unet.set_attn_processor(lora_attn_procs)

    def get_prompt_embeds(
        self,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        negative_input2_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        negative_attention2_mask: Optional[torch.Tensor] = None,
    ):
        prompt_outputs = self.text(
            input_ids,
            # attention_mask,
            output_hidden_states=True,
        )
        prompt_embeds = prompt_outputs.hidden_states[-2]
        negative_prompt_outputs = self.text(
            negative_input_ids,
            # negative_attention_mask,
            output_hidden_states=True,
        )
        negative_prompt_embeds = negative_prompt_outputs.hidden_states[-2]
        prompt2_outputs = self.text2(
            input2_ids,
            # attention2_mask,
            output_hidden_states=True,
        )
        prompt2_embeds = prompt2_outputs.hidden_states[-2]
        negative_prompt2_outputs = self.text2(
            negative_input2_ids,
            # negative_attention2_mask,
            output_hidden_states=True,
        )
        negative_prompt2_embeds = negative_prompt2_outputs.hidden_states[-2]

        prompt_embeds = torch.concat([prompt_embeds, prompt2_embeds], dim=-1)
        negative_prompt_embeds = torch.concat(
            [negative_prompt_embeds, negative_prompt2_embeds], dim=-1
        )
        pooled_prompt_embeds = prompt2_outputs[0]
        negative_pooled_prompt_embeds = negative_prompt2_outputs[0]
        return GenericOutputs(
            prompt_embeds=prompt_embeds,
            negative_prompt_embeds=negative_prompt_embeds,
            pooled_prompt_embeds=pooled_prompt_embeds,
            negative_pooled_prompt_embeds=negative_pooled_prompt_embeds,
        )


class ControlNetXLForText2ImageGeneration(GenericControlNetXLModel):
    """
    ControlNetXL model for text-to-image generation.

    Args:
        config_path (str): Path to the model configuration file.
        text_config_path (str): Path to the text model configuration file.
        text2_config_path (str): Path to the second text model configuration file.
        vae_config_path (str): Path to the VAE model configuration file.
        controlnet_config_path (str): Path to the ControlNet model configuration file.
        scheduler_config_path (str): Path to the scheduler configuration file.
        quant_config_path (Optional[str]): Path to the quantization configuration file (default: None).
        image_size (Optional[int]): Size of the input image (default: None).
        in_channels (Optional[int]): Number of input channels (default: None).
        out_channels (Optional[int]): Number of output channels (default: None).
        num_train_timesteps (Optional[int]): Number of training timesteps (default: 1000).
        num_infer_timesteps (Optional[int]): Number of inference timesteps (default: 50).
        freeze_vae_encoder (Optional[bool]): Whether to freeze the VAE encoder (default: True).
        freeze_text_encoder (Optional[bool]): Whether to freeze the text encoder (default: True).
        freeze_unet_encoder (Optional[bool]): Whether to freeze the UNet encoder (default: True).
        lora_r (Optional[int]): Lora parameter (default: None).
        seed (Optional[int]): Random seed (default: 1123).
    """

    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_unet_encoder: Optional[bool] = True,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            controlnet_config_path=controlnet_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_unet_encoder=freeze_unet_encoder,
            lora_r=lora_r,
            seed=seed,
        )
        self.pipeline = StableDiffusionXLControlNetPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            unet=self.unet,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        add_time_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        condition_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the model.

        Args:
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            add_time_ids (torch.Tensor): Additional time IDs.
            pixel_values (torch.Tensor): Pixel values.
            condition_pixel_values (torch.Tensor): Condition pixel values.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).

        Returns:
            torch.Tensor: Loss value.
        """
        prompt_outputs = self.text(
            input_ids,
            # attention_mask,
            output_hidden_states=True,
        )
        prompt_embeds = prompt_outputs.hidden_states[-2]
        prompt2_outputs = self.text2(
            input2_ids,
            # attention2_mask,
            output_hidden_states=True,
        )
        prompt2_embeds = prompt2_outputs.hidden_states[-2]
        prompt_embeds = torch.concat([prompt_embeds, prompt2_embeds], dim=-1)
        pooled_prompt_embeds = prompt2_outputs[0]

        latents = self.vae.encode(pixel_values).latent_dist.sample()
        latents = latents * self.vae.config.scaling_factor

        noise = torch.randn(latents.shape).to(latents.device)
        batch = latents.size(0)

        timesteps = torch.randint(
            0,
            self.scheduler.config.num_train_timesteps,
            (batch,),
            device=pixel_values.device,
        ).long()

        noise_latents = self.scheduler.add_noise(
            latents,
            noise,
            timesteps,
        )

        encoder_hidden_states = self.text(input_ids, attention_mask)[0]
        down_block_res_samples, mid_block_res_sample = self.controlnet(
            noise_latents,
            timesteps,
            encoder_hidden_states=encoder_hidden_states,
            controlnet_cond=condition_pixel_values,
            added_cond_kwargs={
                "time_ids": add_time_ids,
                "text_embeds": pooled_prompt_embeds,
            },
            return_dict=False,
        )
        outputs = self.unet(
            noise_latents,
            timesteps,
            prompt_embeds,
            added_cond_kwargs={
                "time_ids": add_time_ids,
                "text_embeds": pooled_prompt_embeds,
            },
            down_block_additional_residuals=down_block_res_samples,
            mid_block_additional_residual=mid_block_res_sample,
        ).sample

        if self.scheduler.config.prediction_type == "v_prediction":
            noise = self.scheduler.get_velocity(latents, noise, timesteps)

        loss = F.mse_loss(outputs, noise, reduction="mean")
        return loss

    def generate(
        self,
        condition_pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        negative_input2_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        negative_attention2_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 5.0,
        controlnet_conditioning_scale: Optional[float] = 0.5,
    ):
        """
        Generate images using the model.

        Args:
            condition_pixel_values (torch.Tensor): Condition pixel values.
            input_ids (torch.Tensor): Input IDs.
            input2_ids (torch.Tensor): Second input IDs.
            negative_input_ids (torch.Tensor): Negative input IDs.
            negative_input2_ids (torch.Tensor): Negative second input IDs.
            attention_mask (Optional[torch.Tensor]): Attention mask (default: None).
            attention2_mask (Optional[torch.Tensor]): Second attention mask (default: None).
            negative_attention_mask (Optional[torch.Tensor]): Negative attention mask (default: None).
            negative_attention2_mask (Optional[torch.Tensor]): Negative second attention mask (default: None).
            height (Optional[int]): Height of the generated images (default: 1024).
            width (Optional[int]): Width of the generated images (default: 1024).
            guidance_scale (Optional[float]): Scale for guidance (default: 5.0).

        Returns:
            GenericOutputs: Generated images.
        """
        outputs = self.get_prompt_embeds(
            input_ids=input_ids,
            input2_ids=input2_ids,
            negative_input_ids=negative_input_ids,
            negative_input2_ids=negative_input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
            negative_attention_mask=negative_attention_mask,
            negative_attention2_mask=negative_attention2_mask,
        )
        images = self.pipeline(
            image=condition_pixel_values,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            negative_pooled_prompt_embeds=outputs.negative_pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=float(controlnet_conditioning_scale),
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class ControlNetXLForImage2ImageGeneration(GenericControlNetXLModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_unet_encoder: Optional[bool] = True,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            controlnet_config_path=controlnet_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_unet_encoder=freeze_unet_encoder,
            lora_r=lora_r,
            seed=seed,
        )
        self.pipeline = StableDiffusionXLControlNetImg2ImgPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            unet=self.unet,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        negative_input2_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        condition_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        negative_attention2_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 0.99,
        guidance_scale: Optional[float] = 7.5,
        controlnet_conditioning_scale: Optional[float] = 0.5,
    ):
        outputs = self.get_prompt_embeds(
            input_ids=input_ids,
            input2_ids=input2_ids,
            negative_input_ids=negative_input_ids,
            negative_input2_ids=negative_input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
            negative_attention_mask=negative_attention_mask,
            negative_attention2_mask=negative_attention2_mask,
        )
        images = self.pipeline(
            image=pixel_values,
            control_image=condition_pixel_values,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            negative_pooled_prompt_embeds=outputs.negative_pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            strength=strength,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=float(controlnet_conditioning_scale),
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))


class ControlNetXLForImageInpainting(GenericControlNetXLModel):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        vae_config_path: str,
        controlnet_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        freeze_unet_encoder: Optional[bool] = True,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            vae_config_path=vae_config_path,
            controlnet_config_path=controlnet_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            freeze_unet_encoder=freeze_unet_encoder,
            lora_r=lora_r,
            seed=seed,
        )
        self.pipeline = StableDiffusionXLControlNetInpaintPipeline(
            vae=self.vae,
            text_encoder=self.text,
            text_encoder_2=self.text2,
            unet=self.unet,
            controlnet=self.controlnet,
            scheduler=self.scheduler,
            tokenizer=None,
            tokenizer_2=None,
        )
        self.pipeline.set_progress_bar_config(disable=True)

    def forward(
        self,
    ):
        raise NotImplementedError

    def generate(
        self,
        input_ids: torch.Tensor,
        input2_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        negative_input2_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        pixel_masks: torch.Tensor,
        condition_pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        attention2_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        negative_attention2_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 0.8,
        guidance_scale: Optional[float] = 7.5,
        controlnet_conditioning_scale: Optional[float] = 0.5,
    ):
        outputs = self.get_prompt_embeds(
            input_ids=input_ids,
            input2_ids=input2_ids,
            negative_input_ids=negative_input_ids,
            negative_input2_ids=negative_input2_ids,
            attention_mask=attention_mask,
            attention2_mask=attention2_mask,
            negative_attention_mask=negative_attention_mask,
            negative_attention2_mask=negative_attention2_mask,
        )
        images = self.pipeline(
            image=pixel_values,
            mask_image=pixel_masks,
            control_image=condition_pixel_values,
            prompt_embeds=outputs.prompt_embeds,
            negative_prompt_embeds=outputs.negative_prompt_embeds,
            pooled_prompt_embeds=outputs.pooled_prompt_embeds,
            negative_pooled_prompt_embeds=outputs.negative_pooled_prompt_embeds,
            generator=torch.Generator(device=self.pipeline.device).manual_seed(
                self.seed
            ),
            strength=strength,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=float(controlnet_conditioning_scale),
            output_type="np.array",
        ).images

        return GenericOutputs(images=torch.from_numpy(images))



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from transformers.models.xlm_roberta.modeling_xlm_roberta import (
    XLMRobertaConfig,
    XLMRobertaModel,
)
from transformers.models.xlm_roberta_xl.modeling_xlm_roberta_xl import (
    XLMRobertaXLConfig,
    XLMRobertaXLModel,
)

from transformers.models.roberta.modeling_roberta import RobertaLMHead
from unitorch.models import GenericModel


class XLMRobertaForClassification(GenericModel):
    """
    XLM-RoBERTa model for classification tasks.
    """

    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = 1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes the XLMRobertaForClassification model.

        Args:
            config_path (str): Path to the configuration file.
            num_classes (Optional[int]): Number of classes. Defaults to 1.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = XLMRobertaConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.roberta = XLMRobertaModel(self.config)
        self.dropout = nn.Dropout(self.config.hidden_dropout_prob)
        self.classifier = nn.Linear(self.config.hidden_size, num_classes)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the XLMRobertaForClassification model.

        Args:
            input_ids (torch.Tensor): Input tensor of shape [batch_size, sequence_length].
            attention_mask (Optional[torch.Tensor]): Attention mask tensor of shape [batch_size, sequence_length].
                Defaults to None.
            token_type_ids (Optional[torch.Tensor]): Token type IDs tensor of shape [batch_size, sequence_length].
                Defaults to None.
            position_ids (Optional[torch.Tensor]): Position IDs tensor of shape [batch_size, sequence_length].
                Defaults to None.

        Returns:
            (torch.Tensor):Output logits of shape [batch_size, num_classes].
        """
        outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits


class XLMRobertaForMaskLM(GenericModel):
    """
    XLM-RoBERTa model for masked language modeling tasks.
    """

    def __init__(
        self,
        config_path: str,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes the XLMRobertaForMaskLM model.

        Args:
            config_path (str): Path to the configuration file.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = XLMRobertaConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.roberta = XLMRobertaModel(self.config, add_pooling_layer=False)
        self.lm_head = RobertaLMHead(self.config)
        self.init_weights()
        self.roberta.embeddings.word_embeddings.weight = self.lm_head.decoder.weight

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the XLMRobertaForMaskLM model.

        Args:
            input_ids (torch.Tensor): Input tensor of shape [batch_size, sequence_length].
            attention_mask (Optional[torch.Tensor]): Attention mask tensor of shape [batch_size, sequence_length].
                Defaults to None.
            token_type_ids (Optional[torch.Tensor]): Token type IDs tensor of shape [batch_size, sequence_length].
                Defaults to None.
            position_ids (Optional[torch.Tensor]): Position IDs tensor of shape [batch_size, sequence_length].
                Defaults to None.

        Returns:
            (torch.Tensor):Output logits of shape [batch_size, sequence_length, vocabulary_size].
        """
        outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        sequence_output = outputs[0]
        logits = self.lm_head(sequence_output)
        return logits


class XLMRobertaXLForClassification(GenericModel):
    """
    XLM-RoBERTa XL model for classification tasks.
    """

    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = 1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes the XLMRobertaXLForClassification model.

        Args:
            config_path (str): Path to the configuration file.
            num_classes (Optional[int]): Number of classes for classification. Defaults to 1.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = XLMRobertaXLConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.roberta = XLMRobertaXLModel(self.config)
        self.dropout = nn.Dropout(self.config.hidden_dropout_prob)
        self.classifier = nn.Linear(self.config.hidden_size, num_classes)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """
        Forward pass of the XLMRobertaXLForClassification model.

        Args:
            input_ids (torch.Tensor): Input tensor of shape [batch_size, sequence_length].
            attention_mask (Optional[torch.Tensor]): Attention mask tensor of shape [batch_size, sequence_length].
                Defaults to None.
            token_type_ids (Optional[torch.Tensor]): Token type IDs tensor of shape [batch_size, sequence_length].
                Defaults to None.
            position_ids (Optional[torch.Tensor]): Position IDs tensor of shape [batch_size, sequence_length].
                Defaults to None.

        Returns:
            (torch.Tensor):Output logits of shape [batch_size, num_classes].
        """
        outputs = self.roberta(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
        )
        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from functools import partial

import torch
from transformers import DebertaTokenizer
from unitorch.utils import pop_value, truncate_sequence_pair
from unitorch.models import GenericOutputs, HfTextClassificationProcessor


def get_deberta_tokenizer(vocab_path, merge_path):
    assert os.path.exists(vocab_path) and os.path.exists(merge_path)
    tokenizer = DebertaTokenizer(vocab_path, merge_path)
    return tokenizer


class DebertaProcessor(HfTextClassificationProcessor):
    """
    Processor for DeBERTa-based text classification models.
    """

    def __init__(
        self,
        vocab_path: str,
        merge_path: str,
        max_seq_length: Optional[int] = 128,
        source_type_id: Optional[int] = 0,
        target_type_id: Optional[int] = 0,
    ):
        """
        Initializes the DebertaProcessor.

        Args:
            vocab_path (str): Path to the vocabulary file.
            merge_path (str): Path to the merge file.
            max_seq_length (Optional[int]): Maximum sequence length. Defaults to 128.
            source_type_id (Optional[int]): Source type ID for token types. Defaults to 0.
            target_type_id (Optional[int]): Target type ID for token types. Defaults to 0.
        """
        tokenizer = get_deberta_tokenizer(
            vocab_path,
            merge_path,
        )
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            source_type_id=source_type_id,
            target_type_id=target_type_id,
            position_start_id=tokenizer.pad_token_id + 1,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import time
import json
import logging
import deepspeed
import numpy as np
import pandas as pd
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from copy import deepcopy
from itertools import chain
from collections.abc import Iterable
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union, Iterator
from torch.utils.data import DataLoader, Dataset, SequentialSampler
from torch.utils.data.distributed import DistributedSampler
from torch.cuda.amp import autocast, GradScaler
from torch.multiprocessing import Process, Queue
from unitorch import set_seed, is_torch2_available
from unitorch.models import ExponentialMovingAverage
from unitorch.utils import get_local_rank
from unitorch.utils import (
    DistributedSkipSampler,
    RandomSkipSampler,
    SequentialSkipSampler,
    PostProcess,
    IOProcess,
    GENERATE_FINISHED,
)
from unitorch.models import GenericOutputs
from unitorch.utils import ActiveGPUJob
from unitorch.cli import (
    cached_path,
    register_task,
    registered_model,
    registered_optim,
    registered_dataset,
    registered_loss,
    registered_score,
    registered_scheduler,
    registered_writer,
    init_registered_module,
    init_registered_process,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models import (
    ModelInputs,
    ModelOutputs,
    ModelTargets,
    LossOutputs,
    CombineTensorsInputs,
    CombineTensorsTargets,
)
from unitorch.cli.tasks.supervised import (
    DatasetFeature,
    collate_fn,
    infer,
    monitor,
    save_checkpoint,
)


@register_task("core/task/deepspeed/supervised")
class DeepspeedTask:
    """Task class for deepspeed supervised learning."""

    def __init__(
        self,
        configure,
        model,
        datasets,
        local_rank: Optional[int] = -1,
        seed: Optional[int] = 1123,
    ):
        """
        Initialize the DeepspeedTask.

        Args:
            configure: The configuration object.
            model: The model for supervised learning.
            datasets: The datasets for training and evaluation.
            local_rank (optional): The local rank for distributed training. Defaults to -1.
            seed (optional): The random seed. Defaults to 1123.
        """
        set_seed(seed)
        self.n_gpu = 1 if torch.cuda.is_available() else 0
        if dist.is_initialized():
            self.n_gpu = dist.get_world_size()

        self.config = configure
        self.model = model
        self.datasets = datasets
        self.local_rank = local_rank

        if self.local_rank != -1:
            torch.cuda.set_device(self.local_rank)

        if torch.cuda.is_available():
            self.model = self.model.cuda()

        self.best_score = -np.inf

    @classmethod
    @add_default_section_for_init("core/task/deepspeed/supervised")
    def from_core_configure(cls, config, **kwargs):
        """
        Create a DeepspeedTask instance from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            A DeepspeedTask instance.
        """
        try:
            deepspeed.init_distributed(dist_backend="nccl", init_method="env://")
        except:
            logging.info("PyTorch is not in distributed mode")

        config.set_default_section("core/task/deepspeed/supervised")

        model = config.getoption("model", None)
        dataset = config.getoption("dataset", None)

        if model is not None:
            model = init_registered_module(model, config, registered_model)

        if dataset is not None:
            dataset = init_registered_module(dataset, config, registered_dataset)

        local_rank = config.getdefault(
            "core/cli",
            "local_rank",
            get_local_rank(),
        )

        return dict(
            configure=config,
            model=model,
            datasets=dataset,
            local_rank=local_rank,
        )

    @add_default_section_for_function("core/task/deepspeed/supervised")
    def train(
        self,
        config_path: str,
        optim: str,
        loss_fn: str,
        score_fn: str,
        monitor_fns: Optional[Union[str, List[str]]] = None,
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        to_ckpt_dir: Optional[str] = "./to_ckpt",
        train_batch_size: Optional[int] = 128,
        dev_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
        save_optimizer: Optional[bool] = False,
        save_scheduler: Optional[bool] = False,
        log_freq: Optional[int] = 100,
        ckpt_freq: Optional[int] = 10000,
        grad_acc_step: Optional[int] = 1,
        max_grad_norm: Optional[float] = 1.0,
        learning_rate: Optional[float] = None,
        max_warmup_learning_rate: Optional[float] = None,
        num_warmup_steps: Optional[int] = None,
        epochs: Optional[int] = 5,
        use_amp: Optional[bool] = False,
        use_ema: Optional[bool] = False,
        ema_decay: Optional[float] = 0.9999,
        ema_tau: Optional[int] = 2000,
        gpu_mode: Optional[bool] = False,
    ):
        """
        Train the model using deepspeed.

        Args:
            config_path: The path to the deepspeed configuration file.
            optim: The optimizer used for training.
            loss_fn: The loss function used for training.
            score_fn: The score function used for evaluation.
            monitor_fns (optional): The monitoring functions for evaluation. Defaults to None.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            to_ckpt_dir (optional): The directory path to save checkpoints to. Defaults to "./to_ckpt".
            train_batch_size (optional): The batch size for training. Defaults to 128.
            dev_batch_size (optional): The batch size for evaluation. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
            save_optimizer (optional): Whether to save the optimizer state. Defaults to False.
            save_scheduler (optional): Whether to save the scheduler state. Defaults to False.
            log_freq (optional): The frequency of logging. Defaults to 100.
            ckpt_freq (optional): The frequency of saving checkpoints. Defaults to 10000.
            grad_acc_step (optional): The number of gradient accumulation steps. Defaults to 1.
            max_grad_norm (optional): The maximum gradient norm. Defaults to 1.0.
            learning_rate (optional): The learning rate for the optimizer. Defaults to None.
            max_warmup_learning_rate (optional): The maximum learning rate during warmup. Defaults to None.
            num_warmup_steps (optional): The number of warmup steps. Defaults to None.
            epochs (optional): The number of training epochs. Defaults to 5.
            use_ema (optional): Whether to use exponential moving average. Defaults to False.
            ema_decay (optional): The decay factor for exponential moving average. Defaults to 0.9999.
            ema_tau (optional): The time constant for exponential moving average. Defaults to 2000.
            gpu_mode (optional): Whether to make GPU active. Defaults to False.
        """
        if not os.path.exists(to_ckpt_dir) and self.local_rank in [-1, 0]:
            os.makedirs(to_ckpt_dir, exist_ok=True)

        if loss_fn is not None:
            loss_fn = init_registered_module(loss_fn, self.config, registered_loss)

        if score_fn is not None:
            score_fn = init_registered_module(score_fn, self.config, registered_score)

        if monitor_fns is not None:
            monitor_fns = [
                init_registered_module(monitor_fn, self.config, registered_score)
                for monitor_fn in monitor_fns
                if monitor_fn in registered_score
            ]

        config_file = cached_path(config_path)
        config_dict = json.load(open(config_file, "r"))
        config_dict["train_micro_batch_size_per_gpu"] = train_batch_size

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)

        if os.path.exists(to_ckpt_dir):
            self.model.from_checkpoint(
                to_ckpt_dir,
                weight_name="pytorch_model_latest.bin",
            )

        params = self.model.parameters()
        params = filter(lambda x: x.requires_grad, params)

        assert "optimizer" in config_dict

        if "params" not in config_dict["optimizer"]:
            config_dict["optimizer"]["params"] = dict()

        if "scheduler" in config_dict:
            if "params" not in config_dict["scheduler"]:
                config_dict["scheduler"]["params"] = dict()

        if learning_rate is not None:
            config_dict["optimizer"]["params"]["lr"] = learning_rate
            if "scheduler" in config_dict:
                config_dict["scheduler"]["params"]["warmup_max_lr"] = learning_rate

        if max_warmup_learning_rate is not None and "scheduler" in config_dict:
            config_dict["scheduler"]["params"][
                "warmup_max_lr"
            ] = max_warmup_learning_rate

        if num_warmup_steps is not None and "scheduler" in config_dict:
            if "params" not in config_dict["scheduler"]:
                config_dict["scheduler"]["params"] = dict()
            config_dict["scheduler"]["params"]["warmup_num_steps"] = num_warmup_steps

        info_path = os.path.join(to_ckpt_dir, "info.json")
        if os.path.exists(info_path):
            info = json.load(open(os.path.join(to_ckpt_dir, "info.json")))
        else:
            info = dict()

        global_epoch = info.get("global_epoch", 0)
        global_step = info.get("global_step", 0)
        self.best_score = info.get("best_score", self.best_score)

        logging.info(f"the best score is {self.best_score}")

        self.ema_model = None
        if use_ema:
            num_ema_steps = info.get("num_ema_steps", 0)
            self.ema_model = ExponentialMovingAverage(
                self.model,
                decay=ema_decay,
                tau=ema_tau,
                num_steps=num_ema_steps,
            )
            if os.path.exists(from_ckpt_dir):
                self.ema_model.from_checkpoint(
                    from_ckpt_dir,
                    weight_name="pytorch_ema_model.bin",
                )
            if os.path.exists(to_ckpt_dir):
                self.ema_model.from_checkpoint(
                    to_ckpt_dir,
                    weight_name="pytorch_ema_model_latest.bin",
                )

        for n, p in self.model.named_parameters():
            logging.debug(
                f"{n}: trainable - {p.requires_grad} | tensor shape - {p.shape}"
            )

        self.model, optim, _, scheduler = deepspeed.initialize(
            model=self.model,
            config=config_dict,
            model_parameters=params,
        )

        global_rank = -1
        if self.n_gpu > 1:
            global_rank = dist.get_rank()

        train_sampler = DistributedSkipSampler if self.n_gpu > 1 else RandomSkipSampler
        dev_sampler = DistributedSampler if self.n_gpu > 1 else SequentialSampler

        if gpu_mode:
            with ActiveGPUJob() as _:
                dataset_train = self.datasets.get("train")
                dataset_dev = self.datasets.get("dev")
        else:
            dataset_train = self.datasets.get("train")
            dataset_dev = self.datasets.get("dev")

        iter_train = DataLoader(
            dataset_train,
            sampler=train_sampler(dataset_train)
            if not isinstance(dataset_train, Iterable)
            else None,
            batch_size=train_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        iter_dev = DataLoader(
            dataset_dev,
            sampler=dev_sampler(dataset_dev)
            if not isinstance(dataset_dev, Iterable)
            else None,
            batch_size=dev_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        log_loss = 0
        dev_epoch = 0
        for e in range(0, epochs):
            torch.cuda.empty_cache()
            if e < global_epoch:
                continue

            if hasattr(dataset_train, "set_epoch"):
                dataset_train.set_epoch(e)

            if hasattr(dataset_train, "set_skip_step"):
                dataset_train.set_skip_step(global_step * train_batch_size)

            if hasattr(iter_train.sampler, "set_epoch"):
                iter_train.sampler.set_epoch(e)

            if hasattr(iter_train.sampler, "set_skip_step"):
                iter_train.sampler.set_skip_step(global_step * train_batch_size)

            self.model.train()
            is_update_step = True
            for step, (inputs, targets) in enumerate(iter_train):
                step = step + global_step
                is_update_step = False
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                    targets = targets.cuda()

                if is_torch2_available():
                    with torch.cuda.amp.autocast(
                        enabled=use_amp
                    ) as autocast, torch.backends.cuda.sdp_kernel(
                        enable_flash=False
                    ) as disable:
                        outputs = self.model(**inputs.dict())
                        if isinstance(outputs, LossOutputs):
                            loss = outputs.loss / grad_acc_step
                        else:
                            loss = (
                                loss_fn(outputs=outputs, targets=targets)
                                / grad_acc_step
                            )
                else:
                    with torch.cuda.amp.autocast(enabled=use_amp) as autocast:
                        outputs = self.model(**inputs.dict())
                        if isinstance(outputs, LossOutputs):
                            loss = outputs.loss / grad_acc_step
                        else:
                            loss = (
                                loss_fn(outputs=outputs, targets=targets)
                                / grad_acc_step
                            )

                nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                self.model.backward(loss)

                log_loss += loss.data * grad_acc_step
                if (step + 1) % grad_acc_step == 0:
                    is_update_step = True
                    optim.step()
                    if scheduler is not None:
                        scheduler.step()
                    optim.zero_grad()

                    if use_ema and self.ema_model is not None:
                        self.ema_model.step(
                            self.model.module if self.n_gpu > 1 else self.model
                        )

                if (step + 1) % log_freq == 0 and global_rank in [-1, 0]:
                    logging.info(
                        f"epoch {e} step {step}: loss -- { log_loss / log_freq }"
                    )
                    log_loss = 0

                if (step + 1) % ckpt_freq == 0:
                    if hasattr(dataset_dev, "set_epoch"):
                        dataset_dev.set_epoch(dev_epoch)

                    if hasattr(iter_dev.sampler, "set_epoch"):
                        iter_dev.sampler.set_epoch(dev_epoch)

                    dev_epoch += 1
                    self.best_score = save_checkpoint(
                        self.model.module if self.n_gpu > 1 else self.model,
                        to_ckpt_dir,
                        iter_dev,
                        score_fn,
                        monitor_fns,
                        optim=optim if save_optimizer else None,
                        scheduler=scheduler if save_scheduler else None,
                        ema_model=self.ema_model if use_ema else None,
                        best_score=self.best_score,
                        info_path=info_path,
                        local_rank=self.local_rank,
                        global_epoch=e,
                        global_step=step + 1,
                    )

            if not is_update_step:
                optim.step()
                if scheduler is not None:
                    scheduler.step()
                optim.zero_grad()

                if use_ema and self.ema_model is not None:
                    self.ema_model.step(
                        self.model.module if self.n_gpu > 1 else self.model
                    )

            log_loss = 0

            if hasattr(dataset_dev, "set_epoch"):
                dataset_dev.set_epoch(dev_epoch)

            if hasattr(iter_dev.sampler, "set_epoch"):
                iter_dev.sampler.set_epoch(dev_epoch)

            dev_epoch += 1

            global_step = 0
            self.best_score = save_checkpoint(
                self.model.module if self.n_gpu > 1 else self.model,
                to_ckpt_dir,
                iter_dev,
                score_fn,
                monitor_fns,
                optim=optim if save_optimizer else None,
                scheduler=scheduler if save_scheduler else None,
                ema_model=self.ema_model if use_ema else None,
                best_score=self.best_score,
                info_path=info_path,
                local_rank=self.local_rank,
                global_epoch=e,
                global_step=0,
            )

    @torch.no_grad()
    @add_default_section_for_function("core/task/deepspeed/supervised")
    def eval(
        self,
        monitor_fns: Union[str, List[str]],
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        dev_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
        gpu_mode: Optional[bool] = False,
    ):
        """
        Evaluate the model.

        Args:
            monitor_fns: The monitoring functions for evaluation.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            dev_batch_size (optional): The batch size for evaluation. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
            gpu_mode (optional): Whether to make GPU active. Defaults to False.
        """
        monitor_fns = [
            init_registered_module(monitor_fn, self.config, registered_score)
            for monitor_fn in monitor_fns
            if monitor_fn in registered_score
        ]

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)

        global_rank = -1
        if self.n_gpu > 1:
            self.model = nn.parallel.DistributedDataParallel(
                self.model,
                device_ids=[self.local_rank],
                output_device=self.local_rank,
                find_unused_parameters=False,
                broadcast_buffers=False,
            )
            global_rank = dist.get_rank()

        dev_sampler = DistributedSampler if self.n_gpu > 1 else SequentialSampler
        if gpu_mode:
            with ActiveGPUJob() as _:
                dataset_dev = self.datasets.get("dev")
        else:
            dataset_dev = self.datasets.get("dev")
        iter_dev = DataLoader(
            dataset_dev,
            sampler=dev_sampler(dataset_dev)
            if not isinstance(dataset_dev, Iterable)
            else None,
            batch_size=dev_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        results = infer(self.model.module if self.n_gpu > 1 else self.model, iter_dev)
        if self.local_rank in [-1, 0]:
            monitor(
                outputs=results.outputs,
                targets=results.targets,
                monitor_fns=monitor_fns,
            )

    @torch.no_grad()
    @add_default_section_for_function("core/task/deepspeed/supervised")
    def infer(
        self,
        postprocess_fn: str,
        writer: str,
        test_batch_size: Optional[int] = 128,
        pin_memory: Optional[bool] = True,
        num_workers: Optional[int] = 4,
        max_size: Optional[int] = 10000,
        from_ckpt_dir: Optional[str] = "./from_ckpt",
        output_header: Optional[List] = None,
        output_path: Optional[str] = "./cache/predict.txt",
        postprocess_workers: Optional[int] = 2,
        gpu_mode: Optional[bool] = False,
    ):
        """
        Perform inference using the trained model.

        Args:
            postprocess_fn: The post-processing function for inference.
            writer: The writer for writing the results.
            test_batch_size (optional): The batch size for inference. Defaults to 128.
            pin_memory (optional): Whether to pin memory during data loading. Defaults to True.
            num_workers (optional): The number of worker processes for data loading. Defaults to 4.
            max_size (optional): The maximum size of the dataset for inference. Defaults to 10000.
            from_ckpt_dir (optional): The directory path to load checkpoints from. Defaults to "./from_ckpt".
            output_header (optional): The header for the output file. Defaults to None.
            output_path (optional): The path to save the output file. Defaults to "./cache/predict.txt".
            postprocess_workers (optional): The number of worker processes for post-processing. Defaults to 2.
            gpu_mode (optional): Whether to make GPU active. Defaults to False.
        """
        assert self.n_gpu <= 1
        assert writer is not None

        output_dir = os.path.dirname(output_path)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        if postprocess_fn is not None:
            postprocess_fn = init_registered_process(postprocess_fn, self.config)

        if writer is not None:
            writer = init_registered_module(
                writer,
                self.config,
                registered_writer,
                output_file=output_path,
            )

        skip_step = writer.skip_n_samples

        if os.path.exists(from_ckpt_dir):
            self.model.from_checkpoint(from_ckpt_dir)

        if skip_step == 0:
            sampler = SequentialSampler
        else:
            sampler = SequentialSkipSampler

        if gpu_mode:
            with ActiveGPUJob() as _:
                dataset_test = self.datasets.get("test")
        else:
            dataset_test = self.datasets.get("test")

        iter_test = DataLoader(
            dataset_test,
            sampler=sampler(dataset_test)
            if not isinstance(dataset_test, Iterable)
            else None,
            batch_size=test_batch_size,
            shuffle=False,
            pin_memory=pin_memory,
            num_workers=num_workers,
            collate_fn=collate_fn,
        )

        if skip_step > 0:
            output_file = open(output_path, "a")
        else:
            output_file = open(output_path, "w")

        if skip_step > 0 and hasattr(dataset_test, "set_skip_step"):
            dataset_test.set_skip_step(skip_step)

        if skip_step > 0 and hasattr(iter_test.sampler, "set_skip_step"):
            iter_test.sampler.set_skip_step(skip_step)

        if hasattr(dataset_test, "dataset"):
            data_info = dataset_test.dataset
            data_info = DatasetFeature(data_info)
            iter_data = DataLoader(
                deepcopy(data_info),
                sampler=sampler(data_info)
                if not isinstance(dataset_test, Iterable)
                else None,
                batch_size=test_batch_size,
                shuffle=False,
                pin_memory=pin_memory,
                num_workers=num_workers,
                collate_fn=None,
            )
        else:
            iter_data = None

        if skip_step > 0 and hasattr(iter_data.sampler, "set_skip_step"):
            iter_data.sampler.set_skip_step(skip_step)

        self.model.eval()
        start = time.time()

        data_queue = Queue(maxsize=max_size)
        msg_queue = Queue(maxsize=max_size)
        postprocess_list = []
        for _ in range(postprocess_workers):
            p = PostProcess(
                postprocess_fn,
                data_queue,
                msg_queue,
            )
            postprocess_list.append(p)
            p.start()

        io_process = IOProcess(msg_queue, writer=writer)
        io_process.start()

        if iter_data is None:
            for step, (inputs, _) in enumerate(iter_test):
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                outputs = self.model(**inputs.dict())
                outputs = outputs.cpu()
                data_queue.put((step, outputs))
        else:
            for step, ((inputs, _), _infos) in enumerate(zip(iter_test, iter_data)):
                if torch.cuda.is_available():
                    inputs = inputs.cuda()
                outputs = self.model(**inputs.dict())
                outputs = outputs.cpu()
                if output_header is not None:
                    _infos = {k: _infos[k] for k in output_header if k in _infos}
                    outputs.from_pandas(pd.DataFrame(_infos))
                data_queue.put((step, outputs))

        data_queue.put((-1, GENERATE_FINISHED))
        for p in postprocess_list:
            p.join()

        msg_queue.put((-1, GENERATE_FINISHED))
        io_process.join()

        end = time.time()
        ms = (end - start) * 1000
        logging.info(
            "{:.2f} ms, {:.2f} sample/s".format(
                ms,
                ((len(dataset_test) - skip_step) / ms * 1000),
            )
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.clip.modeling import (
    ClipForPretrain,
    ClipForClassification,
    ClipForTextClassification,
    ClipForImageClassification,
)
from unitorch.models.clip.processing import ClipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import re
import torch
import gc
import gradio as gr
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.models.diffusers import pretrained_diffusers_infos
from unitorch.cli.pipelines.controlnet import (
    ControlNetForText2ImageGenerationPipeline,
    ControlNetForImage2ImageGenerationPipeline,
    ControlNetForImageInpaintingPipeline,
)
from unitorch.cli.webuis import matched_pretrained_names


@register_webui("core/webui/controlnet/text2image")
class ControlNetText2ImageWebUI(GenericWebUI):
    match_patterns = [
        "^stable-v1.5.*controlnet",
        "^stable-v2.*controlnet",
        "^stable-v2.1.*controlnet",
    ]
    block_patterns = [
        ".*blipdiffusion",
        ".*inpainting",
    ]
    pretrained_names = list(pretrained_diffusers_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names, match_patterns, block_patterns
    )
    supported_schedulers = ["DPM++", "DPM++SDE", "UniPC++"]

    def __init__(self, config: CoreConfigureParser):
        self.config = config
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]
        self._iface = gr.Blocks()
        with self._iface:
            with gr.Row(variant="panel"):
                pretrained_name = gr.Dropdown(
                    self.supported_pretrained_names,
                    value=self._name,
                    label="Pretrain Checkpoint Name",
                )
                status = gr.Textbox(label="Model Status", value=self._status)
                click_start = gr.Button(value="Start")
                click_stop = gr.Button(value="Stop")
                click_start.click(
                    self.start, inputs=[pretrained_name], outputs=[status]
                )
                click_stop.click(self.stop, outputs=[status])

            with gr.Row(variant="panel"):
                with gr.Column():
                    condition_image = gr.Image(
                        type="pil", label="Input Condition Image"
                    )
                    prompt = gr.Textbox(label="Input Prompt")
                    negative_prompt = gr.Textbox(label="Input Negative Prompt")
                    height = gr.Slider(512, 1024, value=512, label="Image Height")
                    width = gr.Slider(512, 1024, value=512, label="Image Width")
                    guidance_scale = gr.Slider(
                        0, 10, value=7.5, label="Guidance Scale", step=0.1
                    )
                    controlnet_conditioning_scale = gr.Slider(
                        0,
                        10,
                        value=1.0,
                        label="ControlNet Conditioning Scale",
                        step=0.1,
                    )
                    steps = gr.Slider(
                        0, 1000, value=50, label="Diffusion Steps", step=1
                    )
                    seed = gr.Slider(
                        0, 999999999999, value=42, label="Magic Number", step=1
                    )
                    scheduler = gr.Radio(self.supported_schedulers, label="Sampler")
                    with gr.Row(variant="panel"):
                        freeu_s1 = gr.Slider(
                            0, 10, value=0.9, label="FreeU S1", step=0.1
                        )
                        freeu_s2 = gr.Slider(
                            0, 10, value=0.2, label="FreeU S2", step=0.1
                        )
                        freeu_b1 = gr.Slider(
                            0, 10, value=1.2, label="FreeU B1", step=0.1
                        )
                        freeu_b2 = gr.Slider(
                            0, 10, value=1.4, label="FreeU B2", step=0.1
                        )

                    submit = gr.Button(value="Submit")

                image = gr.Image(type="pil", label="Output Image")
                submit.click(
                    self.serve,
                    inputs=[
                        prompt,
                        condition_image,
                        negative_prompt,
                        height,
                        width,
                        guidance_scale,
                        controlnet_conditioning_scale,
                        steps,
                        seed,
                        scheduler,
                        freeu_s1,
                        freeu_s2,
                        freeu_b1,
                        freeu_b2,
                    ],
                    outputs=[image],
                )

            self._iface.load(
                fn=lambda: gr.update(value=self._name), outputs=[pretrained_name]
            )
            self._iface.load(fn=lambda: gr.update(value=self._status), outputs=[status])

    @property
    def name(self):
        return "ControlNetText2Image"

    @property
    def iface(self):
        return self._iface

    @property
    def status(self):
        return self._status == "running"

    def start(self, pretrained_name, **kwargs):
        if self._status == "running":
            self.stop()
        self.config.set(
            "core/pipeline/controlnet/text2image", "pretrained_name", pretrained_name
        )
        if pretrained_name.startswith("stable-v2"):
            self.config.set("core/pipeline/controlnet/text2image", "pad_token", "!")
        self._name = pretrained_name
        self._pipe = ControlNetForText2ImageGenerationPipeline.from_core_configure(
            self.config
        )
        self._status = "running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        return self._status

    def serve(
        self,
        text: str,
        condition_image: Image.Image,
        neg_text: Optional[str] = "",
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        guidance_scale: Optional[float] = 7.5,
        controlnet_conditioning_scale: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
        scheduler: Optional[str] = None,
        freeu_s1: Optional[float] = 0.9,
        freeu_s2: Optional[float] = 0.2,
        freeu_b1: Optional[float] = 1.2,
        freeu_b2: Optional[float] = 1.4,
    ):
        assert self._pipe is not None
        image = self._pipe(
            text,
            condition_image=condition_image,
            neg_text=neg_text,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            num_timesteps=num_timesteps,
            seed=seed,
            scheduler=scheduler,
            freeu_params=(freeu_s1, freeu_s2, freeu_b1, freeu_b2),
        )
        return image


@register_webui("core/webui/controlnet/image2image")
class ControlNetImage2ImageWebUI(GenericWebUI):
    match_patterns = [
        "^stable-v1.5.*controlnet",
        "^stable-v2.*controlnet",
        "^stable-v2.1.*controlnet",
    ]
    block_patterns = [
        ".*blipdiffusion",
    ]
    pretrained_names = list(pretrained_diffusers_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names, match_patterns, block_patterns
    )
    supported_schedulers = ["DPM++", "DPM++SDE", "UniPC++"]

    def __init__(self, config: CoreConfigureParser):
        self.config = config
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]
        self._iface = gr.Blocks()
        with self._iface:
            with gr.Row(variant="panel"):
                pretrained_name = gr.Dropdown(
                    self.supported_pretrained_names,
                    value=self._name,
                    label="Pretrain Checkpoint Name",
                )
                status = gr.Textbox(label="Model Status", value=self._status)
                click_start = gr.Button(value="Start")
                click_stop = gr.Button(value="Stop")
                click_start.click(
                    self.start, inputs=[pretrained_name], outputs=[status]
                )
                click_stop.click(self.stop, outputs=[status])

            with gr.Row(variant="panel"):
                with gr.Column():
                    with gr.Row():
                        raw_image = gr.Image(type="pil", label="Input Image")
                        condition_image = gr.Image(
                            type="pil", label="Input Condition Image"
                        )
                    prompt = gr.Textbox(label="Input Prompt")
                    negative_prompt = gr.Textbox(label="Input Negative Prompt")
                    strength = gr.Slider(0, 1, value=0.8, label="Strength", step=0.01)
                    guidance_scale = gr.Slider(
                        0, 10, value=7.5, label="Guidance Scale", step=0.1
                    )
                    controlnet_conditioning_scale = gr.Slider(
                        0,
                        10,
                        value=1.0,
                        label="ControlNet Conditioning Scale",
                        step=0.1,
                    )
                    steps = gr.Slider(
                        0, 1000, value=50, label="Diffusion Steps", step=1
                    )
                    seed = gr.Slider(
                        0, 999999999999, value=42, label="Magic Number", step=1
                    )
                    scheduler = gr.Radio(self.supported_schedulers, label="Sampler")
                    with gr.Row(variant="panel"):
                        freeu_s1 = gr.Slider(
                            0, 10, value=0.9, label="FreeU S1", step=0.1
                        )
                        freeu_s2 = gr.Slider(
                            0, 10, value=0.2, label="FreeU S2", step=0.1
                        )
                        freeu_b1 = gr.Slider(
                            0, 10, value=1.2, label="FreeU B1", step=0.1
                        )
                        freeu_b2 = gr.Slider(
                            0, 10, value=1.4, label="FreeU B2", step=0.1
                        )

                    submit = gr.Button(value="Submit")
                image = gr.Image(type="pil", label="Output Image")
                submit.click(
                    self.serve,
                    inputs=[
                        prompt,
                        raw_image,
                        condition_image,
                        negative_prompt,
                        strength,
                        guidance_scale,
                        controlnet_conditioning_scale,
                        steps,
                        seed,
                        scheduler,
                        freeu_s1,
                        freeu_s2,
                        freeu_b1,
                        freeu_b2,
                    ],
                    outputs=[image],
                )

            self._iface.load(
                fn=lambda: gr.update(value=self._name), outputs=[pretrained_name]
            )
            self._iface.load(fn=lambda: gr.update(value=self._status), outputs=[status])

    @property
    def name(self):
        return "ControlNetImage2Image"

    @property
    def iface(self):
        return self._iface

    @property
    def status(self):
        return self._status == "running"

    def start(self, pretrained_name, **kwargs):
        if self._status == "running":
            self.stop()
        self.config.set(
            "core/pipeline/controlnet/image2image", "pretrained_name", pretrained_name
        )
        if pretrained_name.startswith("stable-v2"):
            self.config.set("core/pipeline/controlnet/image2image", "pad_token", "!")
        self._name = pretrained_name
        self._pipe = ControlNetForImage2ImageGenerationPipeline.from_core_configure(
            self.config
        )
        self._status = "running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        return self._status

    def serve(
        self,
        text: str,
        image: Image.Image,
        condition_image: Image.Image,
        neg_text: Optional[str] = "",
        strength: Optional[float] = 0.8,
        guidance_scale: Optional[float] = 7.5,
        controlnet_conditioning_scale: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
        scheduler: Optional[str] = None,
        freeu_s1: Optional[float] = 0.9,
        freeu_s2: Optional[float] = 0.2,
        freeu_b1: Optional[float] = 1.2,
        freeu_b2: Optional[float] = 1.4,
    ):
        assert self._pipe is not None
        image = self._pipe(
            text,
            image,
            condition_image,
            neg_text=neg_text,
            strength=strength,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            num_timesteps=num_timesteps,
            seed=seed,
            scheduler=scheduler,
            freeu_params=(freeu_s1, freeu_s2, freeu_b1, freeu_b2),
        )
        return image


@register_webui("core/webui/controlnet/inpainting")
class ControlNetImageInpaintingWebUI(GenericWebUI):
    match_patterns = [
        "^stable-v1.5.*controlnet.*inpainting",
        "^stable-v2.*controlnet.*inpainting",
        "^stable-v2.1.*controlnet.*inpainting",
    ]
    block_patterns = [
        ".*blipdiffusion",
    ]
    pretrained_names = list(pretrained_diffusers_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names, match_patterns, block_patterns
    )
    supported_schedulers = ["DPM++", "DPM++SDE", "UniPC++"]

    def __init__(self, config: CoreConfigureParser):
        self.config = config
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]
        self._iface = gr.Blocks()
        with self._iface:
            with gr.Row(variant="panel"):
                pretrained_name = gr.Dropdown(
                    self.supported_pretrained_names,
                    value=self._name,
                    label="Pretrain Checkpoint Name",
                )
                status = gr.Textbox(label="Model Status", value=self._status)
                click_start = gr.Button(value="Start")
                click_stop = gr.Button(value="Stop")
                click_start.click(
                    self.start, inputs=[pretrained_name], outputs=[status]
                )
                click_stop.click(self.stop, outputs=[status])

            with gr.Row(variant="panel"):
                with gr.Column():
                    with gr.Row():
                        raw_image = gr.Image(type="pil", label="Input Image")
                        raw_image_mask = gr.Image(type="pil", label="Input Mask Image")
                        condition_image = gr.Image(
                            type="pil", label="Input Condition Image"
                        )

                    prompt = gr.Textbox(label="Input Prompt")
                    negative_prompt = gr.Textbox(label="Input Negative Prompt")
                    strength = gr.Slider(0, 1, value=0.8, label="Strength", step=0.01)
                    guidance_scale = gr.Slider(
                        0, 10, value=7.5, label="Guidance Scale", step=0.1
                    )
                    controlnet_conditioning_scale = gr.Slider(
                        0,
                        10,
                        value=1.0,
                        label="ControlNet Conditioning Scale",
                        step=0.1,
                    )
                    steps = gr.Slider(
                        0, 1000, value=50, label="Diffusion Steps", step=1
                    )
                    seed = gr.Slider(
                        0, 999999999999, value=42, label="Magic Number", step=1
                    )
                    scheduler = gr.Radio(self.supported_schedulers, label="Sampler")
                    with gr.Row(variant="panel"):
                        freeu_s1 = gr.Slider(
                            0, 10, value=0.9, label="FreeU S1", step=0.1
                        )
                        freeu_s2 = gr.Slider(
                            0, 10, value=0.2, label="FreeU S2", step=0.1
                        )
                        freeu_b1 = gr.Slider(
                            0, 10, value=1.2, label="FreeU B1", step=0.1
                        )
                        freeu_b2 = gr.Slider(
                            0, 10, value=1.4, label="FreeU B2", step=0.1
                        )

                    submit = gr.Button(value="Submit")
                image = gr.Image(type="pil", label="Output Image")

                submit.click(
                    self.serve,
                    inputs=[
                        prompt,
                        raw_image,
                        raw_image_mask,
                        condition_image,
                        negative_prompt,
                        strength,
                        guidance_scale,
                        controlnet_conditioning_scale,
                        steps,
                        seed,
                        scheduler,
                        freeu_s1,
                        freeu_s2,
                        freeu_b1,
                        freeu_b2,
                    ],
                    outputs=[image],
                )

            self._iface.load(
                fn=lambda: gr.update(value=self._name), outputs=[pretrained_name]
            )
            self._iface.load(fn=lambda: gr.update(value=self._status), outputs=[status])

    @property
    def name(self):
        return "ControlNetImageInpainting"

    @property
    def iface(self):
        return self._iface

    @property
    def status(self):
        return self._status == "running"

    def start(self, pretrained_name, **kwargs):
        if self._status == "running":
            self.stop()
        self.config.set(
            "core/pipeline/controlnet/inpainting", "pretrained_name", pretrained_name
        )
        if pretrained_name.startswith("stable-v2"):
            self.config.set("core/pipeline/controlnet/inpainting", "pad_token", "!")
        self._name = pretrained_name
        self._pipe = ControlNetForImageInpaintingPipeline.from_core_configure(
            self.config
        )
        self._status = "running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "stopped" if self._pipe is None else "running"
        return self._status

    def serve(
        self,
        text: str,
        image: Image.Image,
        mask_image: Image.Image,
        condition_image: Image.Image,
        neg_text: Optional[str] = "",
        strength: Optional[float] = 0.8,
        guidance_scale: Optional[float] = 7.5,
        controlnet_conditioning_scale: Optional[float] = 1.0,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
        scheduler: Optional[str] = None,
        freeu_s1: Optional[float] = 0.9,
        freeu_s2: Optional[float] = 0.2,
        freeu_b1: Optional[float] = 1.2,
        freeu_b2: Optional[float] = 1.4,
    ):
        assert self._pipe is not None
        image = self._pipe(
            text,
            image,
            mask_image,
            condition_image,
            neg_text=neg_text,
            strength=strength,
            guidance_scale=guidance_scale,
            controlnet_conditioning_scale=controlnet_conditioning_scale,
            num_timesteps=num_timesteps,
            seed=seed,
            scheduler=scheduler,
            freeu_params=(freeu_s1, freeu_s2, freeu_b1, freeu_b2),
        )
        return image



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import json
import math
import random
import logging
import transformers
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as dist
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from transformers import LlamaConfig, LlamaForCausalLM
from transformers.models.blip_2.modeling_blip_2 import (
    Blip2Config,
    Blip2VisionModel,
    Blip2QFormerModel,
)
from unitorch.utils.decorators import replace
from unitorch.models import (
    GenericModel,
    GenericOutputs,
    QuantizationConfig,
    QuantizationMixin,
)
from unitorch.models.clip.modeling import AllGather


class MiniGPT4Blip2LlamaModel(nn.Module):
    """
    MiniGPT4Blip2LlamaModel is a model that combines the Blip2VisionModel, Blip2QFormerModel, and LlamaForCausalLM
    models for generation. It inherits from the nn.Module class.
    """

    def __init__(
        self,
        blip2_config: Blip2Config,
        llama_config: LlamaConfig,
    ):
        """
        Initializes a MiniGPT4Blip2LlamaModel instance.

        Args:
            blip2_config (Blip2Config): The configuration for the Blip2 model.
            llama_config (LlamaConfig): The configuration for the Llama model.
        """
        super().__init__()
        self.blip2_config = blip2_config
        self.vision_model = Blip2VisionModel(self.blip2_config.vision_config)

        self.query_tokens = nn.Parameter(
            torch.zeros(
                1,
                self.blip2_config.num_query_tokens,
                self.blip2_config.qformer_config.hidden_size,
            )
        )
        self.qformer = Blip2QFormerModel(self.blip2_config.qformer_config)

        self.llama_config = llama_config
        self.language_projection = nn.Linear(
            self.blip2_config.qformer_config.hidden_size, self.llama_config.hidden_size
        )
        self.llama = LlamaForCausalLM(self.llama_config)

    def forward(
        self,
        pixel_values: torch.Tensor,
        prefix_input_ids: torch.Tensor,
        suffix_input_ids: torch.Tensor,
        decoder_input_ids: torch.Tensor,
        prefix_attention_mask: Optional[torch.Tensor] = None,
        suffix_attention_mask: Optional[torch.Tensor] = None,
        decoder_attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Performs a forward pass of the MiniGPT4Blip2LlamaModel.

        Args:
            pixel_values (torch.Tensor): The pixel values.
            prefix_input_ids (torch.Tensor): The input IDs for the prefix tokens.
            suffix_input_ids (torch.Tensor): The input IDs for the suffix tokens.
            decoder_input_ids (torch.Tensor): The input IDs for the decoder tokens.
            prefix_attention_mask (torch.Tensor, optional): The attention mask for the prefix tokens.
            suffix_attention_mask (torch.Tensor, optional): The attention mask for the suffix tokens.
            decoder_attention_mask (torch.Tensor, optional): The attention mask for the decoder tokens.

        Returns:
            outputs: The model outputs.
        """
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
        )
        image_embeds = vision_outputs[0]
        image_attention_mask = torch.ones(
            image_embeds.size()[:-1], dtype=torch.long, device=image_embeds.device
        )

        query_tokens = self.query_tokens.expand(image_embeds.shape[0], -1, -1)
        query_outputs = self.qformer(
            query_embeds=query_tokens,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=image_attention_mask,
        )
        query_embeds = query_outputs[0]

        language_model_inputs = self.language_projection(query_embeds)
        language_model_attention_mask = torch.ones(
            language_model_inputs.size()[:-1],
            dtype=torch.long,
            device=language_model_inputs.device,
        )
        prefix_inputs_embeds = self.llama.get_input_embeddings()(prefix_input_ids)
        suffix_inputs_embeds = self.llama.get_input_embeddings()(suffix_input_ids)
        decoder_input_embeds = self.llama.get_input_embeddings()(decoder_input_ids)
        inputs_embeds = torch.cat(
            [
                prefix_inputs_embeds,
                language_model_inputs,
                suffix_inputs_embeds,
                decoder_input_embeds,
            ],
            dim=1,
        )
        expected_device = language_model_attention_mask.device

        if prefix_attention_mask is None:
            prefix_attention_mask = torch.ones(
                prefix_inputs_embeds.size()[:-1],
                dtype=torch.long,
                device=expected_device,
            )

        if suffix_attention_mask is None:
            suffix_attention_mask = torch.ones(
                suffix_inputs_embeds.size()[:-1],
                dtype=torch.long,
                device=expected_device,
            )

        if decoder_attention_mask is None:
            decoder_attention_mask = torch.ones(
                decoder_input_embeds.size()[:-1],
                dtype=torch.long,
                device=expected_device,
            )

        attention_mask = torch.cat(
            [
                prefix_attention_mask,
                language_model_attention_mask,
                suffix_attention_mask,
                decoder_attention_mask.to(expected_device),
            ],
            dim=1,
        )
        outputs = self.llama(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            return_dict=True,
        )
        return outputs

    def generate(
        self,
        pixel_values: torch.FloatTensor,
        prefix_input_ids: Optional[torch.Tensor] = None,
        suffix_input_ids: Optional[torch.Tensor] = None,
        **generate_kwargs,
    ):
        """
        Generates sequences using the MiniGPT4Blip2LlamaModel.

        Args:
            pixel_values (torch.FloatTensor): The pixel values.
            prefix_input_ids (torch.Tensor, optional): The input IDs for the prefix tokens. Defaults to None.
            suffix_input_ids (torch.Tensor, optional): The input IDs for the suffix tokens. Defaults to None.
            **generate_kwargs: Additional keyword arguments for sequence generation.

        Returns:
            outputs: The generation outputs.
        """
        vision_outputs = self.vision_model(
            pixel_values=pixel_values,
        )
        image_embeds = vision_outputs[0]
        image_attention_mask = torch.ones(
            image_embeds.size()[:-1], dtype=torch.long, device=image_embeds.device
        )

        query_tokens = self.query_tokens.expand(image_embeds.shape[0], -1, -1)
        query_outputs = self.qformer(
            query_embeds=query_tokens,
            encoder_hidden_states=image_embeds,
            encoder_attention_mask=image_attention_mask,
        )
        query_embeds = query_outputs[0]

        inputs_embeds = self.language_projection(query_embeds)

        attention_mask = torch.ones(
            inputs_embeds.size(0), inputs_embeds.size(1), dtype=torch.bool
        ).to(inputs_embeds.device)

        if prefix_input_ids is not None:
            prefix_inputs_embeds = self.llama.get_input_embeddings()(prefix_input_ids)
            inputs_embeds = torch.cat([prefix_inputs_embeds, inputs_embeds], dim=1)
            attention_mask = torch.cat(
                [prefix_input_ids.ne(self.blip2_config.pad_token_id), attention_mask],
                dim=1,
            )

        if suffix_input_ids is not None:
            suffix_inputs_embeds = self.llama.get_input_embeddings()(suffix_input_ids)
            inputs_embeds = torch.cat([inputs_embeds, suffix_inputs_embeds], dim=1)
            attention_mask = torch.cat(
                [attention_mask, suffix_input_ids.ne(self.blip2_config.pad_token_id)],
                dim=1,
            )

        outputs = self.llama.generate(
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            **generate_kwargs,
        )

        return outputs


class MiniGPT4Blip2LlamaForGeneration(GenericModel, QuantizationMixin):
    """
    MiniGPT4Blip2LlamaForGeneration is a generation model that combines the MiniGPT4Blip2LlamaModel with generation
    capabilities. It inherits from the GenericModel class.
    """

    prefix_keys_in_state_dict = {
        "^qformer.": "base_model.",
        "^query_tokens": "base_model.",
        "^vision_model.": "base_model.",
        "^model.language_projection.": "base_",
        "^(?!model\.language_projection\.)model\.": "base_model.llama.",
        "^lm_head.": "base_model.llama.",
    }

    def __init__(
        self,
        blip2_config_path: str,
        llama_config_path: str,
        quant_config_path: Optional[str] = None,
        pad_token_id: Optional[int] = 0,
        freeze_vision_model: Optional[bool] = True,
        freeze_qformer_model: Optional[bool] = True,
        freeze_llama_model: Optional[bool] = True,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a MiniGPT4Blip2LlamaForGeneration instance.

        Args:
            blip2_config_path (str): The path to the Blip2 model configuration file.
            llama_config_path (str): The path to the Llama model configuration file.
            pad_token_id (int, optional): The ID of the padding token. Defaults to 0.
            freeze_vision_model (bool, optional): Whether to freeze the parameters of the vision model. Defaults to True.
            freeze_qformer_model (bool, optional): Whether to freeze the parameters of the qformer model. Defaults to True.
            freeze_llama_model (bool, optional): Whether to freeze the parameters of the llama model. Defaults to True.
            gradient_checkpointing (bool, optional): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.blip2_config = Blip2Config.from_json_file(blip2_config_path)
        self.blip2_config.pad_token_id = pad_token_id
        self.llama_config = LlamaConfig.from_json_file(llama_config_path)
        self.llama_config.gradient_checkpointing = gradient_checkpointing
        self.base_model = MiniGPT4Blip2LlamaModel(self.blip2_config, self.llama_config)
        self.init_weights()

        if freeze_vision_model:
            for param in self.base_model.vision_model.parameters():
                param.requires_grad = False

        if freeze_qformer_model:
            for param in self.base_model.qformer.parameters():
                param.requires_grad = False

        if freeze_llama_model:
            for param in self.base_model.llama.parameters():
                param.requires_grad = False

        if quant_config_path is not None:
            self.quant_config = QuantizationConfig.from_json_file(quant_config_path)
            self.quantize(
                self.quant_config, ignore_modules=["language_projection", "lm_head"]
            )

    def forward(
        self,
        pixel_values: torch.Tensor,
        prefix_input_ids: torch.Tensor,
        suffix_input_ids: torch.Tensor,
        decoder_input_ids: torch.Tensor,
        prefix_attention_mask: Optional[torch.Tensor] = None,
        suffix_attention_mask: Optional[torch.Tensor] = None,
        decoder_attention_mask: Optional[torch.Tensor] = None,
    ):
        """
        Performs a forward pass of the MiniGPT4Blip2LlamaForGeneration model.

        Args:
            pixel_values (torch.Tensor): The pixel values.
            prefix_input_ids (torch.Tensor): The input IDs for the prefix tokens.
            suffix_input_ids (torch.Tensor): The input IDs for the suffix tokens.
            decoder_input_ids (torch.Tensor): The input IDs for the decoder tokens.
            prefix_attention_mask (torch.Tensor, optional): The attention mask for the prefix tokens.
            suffix_attention_mask (torch.Tensor, optional): The attention mask for the suffix tokens.
            decoder_attention_mask (torch.Tensor, optional): The attention mask for the decoder tokens.

        Returns:
            logits: The output logits.
        """
        outputs = self.base_model(
            pixel_values=pixel_values,
            prefix_input_ids=prefix_input_ids,
            suffix_input_ids=suffix_input_ids,
            decoder_input_ids=decoder_input_ids,
            prefix_attention_mask=prefix_attention_mask,
            suffix_attention_mask=suffix_attention_mask,
            decoder_attention_mask=decoder_attention_mask,
        )
        logits = outputs.logits[
            :, -suffix_input_ids.size(1) - decoder_input_ids.size(1) :, :
        ]
        return logits

    @torch.no_grad()
    def generate(
        self,
        pixel_values: torch.Tensor,
        prefix_input_ids: torch.Tensor,
        suffix_input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 1,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 2,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 48,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        """
        Generates sequences using the MiniGPT4Blip2LlamaForGeneration model.

        Args:
            pixel_values (torch.Tensor): The pixel values.
            prefix_input_ids (torch.Tensor): The input IDs for the prefix tokens.
            suffix_input_ids (torch.Tensor): The input IDs for the suffix tokens.
            num_beams (int, optional): The number of beams for beam search. Defaults to 5.
            decoder_start_token_id (int, optional): The ID of the decoder start token. Defaults to 1.
            decoder_end_token_id (int or List[int], optional): The ID(s) of the decoder end token(s). Defaults to 2.
            num_return_sequences (int, optional): The number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (int, optional): The minimum length of the generated sequences. Defaults to 0.
            max_gen_seq_length (int, optional): The maximum length of the generated sequences. Defaults to 48.
            repetition_penalty (float, optional): The repetition penalty. Defaults to 1.0.
            no_repeat_ngram_size (int, optional): The size of the n-grams to avoid repeating. Defaults to 0.
            early_stopping (bool, optional): Whether to stop generation early when all beams are finished. Defaults to True.
            length_penalty (float, optional): The length penalty. Defaults to 1.0.
            num_beam_groups (int, optional): The number of groups for diverse beam search. Defaults to 1.
            diversity_penalty (float, optional): The diversity penalty. Defaults to 0.0.
            do_sample (bool, optional): Whether to use sampling for generation. Defaults to False.
            temperature (float, optional): The temperature value for sampling. Defaults to 1.0.
            top_k (int, optional): The value for top-k sampling. Defaults to 50.
            top_p (float, optional): The value for top-p sampling. Defaults to 1.0.

        Returns:
            outputs (GenericOutputs): The generated sequences and their scores.
        """
        outputs = self.base_model.generate(
            pixel_values=pixel_values,
            prefix_input_ids=prefix_input_ids,
            suffix_input_ids=suffix_input_ids,
            max_length=max_gen_seq_length,
            min_length=min_gen_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            decoder_start_token_id=decoder_start_token_id,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )

        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = torch.zeros(
            sequences.size(0), num_return_sequences, max_gen_seq_length
        ).to(device=sequences.device)
        outputs.sequences[:, :, : sequences.size(-1)].copy_(
            sequences[:, :, : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_blip_infos = {
    "default-blip": {
        "config": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/config.json",
        "vision_config": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/vocab.txt",
    },
    "blip-vqa-base": {
        "config": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/config.json",
        "vision_config": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/Salesforce/blip-vqa-base/resolve/main/pytorch_model.bin",
    },
    "blip-image-captioning-base": {
        "config": "https://huggingface.co/Salesforce/blip-image-captioning-base/resolve/main/config.json",
        "vision_config": "https://huggingface.co/Salesforce/blip-image-captioning-base/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/Salesforce/blip-image-captioning-base/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/Salesforce/blip-image-captioning-base/resolve/main/pytorch_model.bin",
    },
    "blip-image-captioning-large": {
        "config": "https://huggingface.co/Salesforce/blip-image-captioning-large/resolve/main/config.json",
        "vision_config": "https://huggingface.co/Salesforce/blip-image-captioning-large/resolve/main/preprocessor_config.json",
        "vocab": "https://huggingface.co/Salesforce/blip-image-captioning-large/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/Salesforce/blip-image-captioning-large/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.blip.modeling
import unitorch.cli.models.blip.processing
from unitorch.cli.models.blip.modeling import (
    BlipForPretrain,
    BlipForClassification,
    BlipForTextClassification,
    BlipForImageClassification,
    BlipForImageCaption,
)
from unitorch.cli.models.blip.processing import BlipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch.cuda.amp import autocast

from unitorch.models.diffusers import (
    StableForText2ImageGeneration as _StableForText2ImageGeneration,
    StableForImage2ImageGeneration as _StableForImage2ImageGeneration,
    StableForImageInpainting as _StableForImageInpainting,
    StableForImageResolution as _StableForImageResolution,
)
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import DiffusionOutputs, LossOutputs
from unitorch.cli.models import diffusion_model_decorator
from unitorch.cli.models.diffusers import pretrained_diffusers_infos, load_weight


@register_model("core/model/diffusers/text2image/stable", diffusion_model_decorator)
class StableForText2ImageGeneration(_StableForText2ImageGeneration):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        snr_gamma: Optional[float] = 5.0,
        lora_r: Optional[int] = None,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            snr_gamma=snr_gamma,
            lora_r=lora_r,
            seed=seed,
        )

    @classmethod
    @add_default_section_for_init("core/model/diffusers/text2image/stable")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/diffusers/text2image/stable")
        pretrained_name = config.getoption("pretrained_name", "stable-v1.5")
        pretrain_infos = nested_dict_value(pretrained_diffusers_infos, pretrained_name)

        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrain_infos, "unet", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = config.getoption("text_config_path", None)
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrain_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        vae_config_path = config.getoption("vae_config_path", None)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrain_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = config.getoption("scheduler_config_path", None)
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrain_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        quant_config_path = config.getoption("quant_config_path", None)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        image_size = config.getoption("image_size", None)
        in_channels = config.getoption("in_channels", None)
        out_channels = config.getoption("out_channels", None)
        num_train_timesteps = config.getoption("num_train_timesteps", 1000)
        num_infer_timesteps = config.getoption("num_infer_timesteps", 50)
        freeze_vae_encoder = config.getoption("freeze_vae_encoder", True)
        freeze_text_encoder = config.getoption("freeze_text_encoder", True)
        snr_gamma = config.getoption("snr_gamma", 5.0)
        lora_r = config.getoption("lora_r", None)
        seed = config.getoption("seed", 1123)

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            snr_gamma=snr_gamma,
            lora_r=lora_r,
            seed=seed,
        )

        weight_path = config.getoption("pretrained_weight_path", None)

        state_dict = None
        if weight_path is None and pretrain_infos is not None:
            state_dict = [
                load_weight(nested_dict_value(pretrain_infos, "unet", "weight")),
                load_weight(nested_dict_value(pretrain_infos, "text", "weight")),
                load_weight(nested_dict_value(pretrain_infos, "vae", "weight")),
            ]

        inst.from_pretrained(weight_path, state_dict=state_dict)
        return inst

    @autocast()
    def forward(
        self,
        pixel_values: torch.Tensor,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ):
        loss = super().forward(
            pixel_values=pixel_values,
            input_ids=input_ids,
            attention_mask=attention_mask,
        )
        return LossOutputs(loss=loss)

    @add_default_section_for_function("core/model/diffusers/text2image/stable")
    @autocast()
    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 7.5,
    ):
        outputs = super().generate(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
            height=height,
            width=width,
            guidance_scale=guidance_scale,
        )

        return DiffusionOutputs(outputs=outputs.images)


@register_model("core/model/diffusers/image2image/stable", diffusion_model_decorator)
class StableForImage2ImageGeneration(_StableForImage2ImageGeneration):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

    @classmethod
    @add_default_section_for_init("core/model/diffusers/image2image/stable")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/diffusers/image2image/stable")
        pretrained_name = config.getoption(
            "pretrained_name", "stable-v1.5-nitrosocke-ghibli"
        )
        pretrain_infos = nested_dict_value(pretrained_diffusers_infos, pretrained_name)

        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrain_infos, "unet", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = config.getoption("text_config_path", None)
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrain_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        vae_config_path = config.getoption("vae_config_path", None)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrain_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = config.getoption("scheduler_config_path", None)
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrain_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        quant_config_path = config.getoption("quant_config_path", None)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        image_size = config.getoption("image_size", None)
        in_channels = config.getoption("in_channels", None)
        out_channels = config.getoption("out_channels", None)
        num_train_timesteps = config.getoption("num_train_timesteps", 1000)
        num_infer_timesteps = config.getoption("num_infer_timesteps", 50)
        freeze_vae_encoder = config.getoption("freeze_vae_encoder", True)
        freeze_text_encoder = config.getoption("freeze_text_encoder", True)
        seed = config.getoption("seed", 1123)

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

        weight_path = config.getoption("pretrained_weight_path", None)

        if weight_path is None and pretrain_infos is not None:
            weight_path = [
                cached_path(nested_dict_value(pretrain_infos, "unet", "weight")),
                cached_path(nested_dict_value(pretrain_infos, "text", "weight")),
                cached_path(nested_dict_value(pretrain_infos, "vae", "weight")),
            ]

        inst.from_pretrained(weight_path)
        return inst

    @autocast()
    def forward(
        self,
    ):
        raise NotImplementedError

    @add_default_section_for_function("core/model/diffusers/image2image/stable")
    @autocast()
    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 0.8,
        guidance_scale: Optional[float] = 7.5,
    ):
        outputs = super().generate(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
            strength=strength,
            guidance_scale=guidance_scale,
        )

        return DiffusionOutputs(outputs=outputs.images)


@register_model("core/model/diffusers/inpainting/stable", diffusion_model_decorator)
class StableForImageInpainting(_StableForImageInpainting):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

    @classmethod
    @add_default_section_for_init("core/model/diffusers/inpainting/stable")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/diffusers/inpainting/stable")
        pretrained_name = config.getoption("pretrained_name", "stable-v1.5-inpainting")
        pretrain_infos = nested_dict_value(pretrained_diffusers_infos, pretrained_name)

        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrain_infos, "unet", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = config.getoption("text_config_path", None)
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrain_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        vae_config_path = config.getoption("vae_config_path", None)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrain_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = config.getoption("scheduler_config_path", None)
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrain_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        quant_config_path = config.getoption("quant_config_path", None)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        image_size = config.getoption("image_size", None)
        in_channels = config.getoption("in_channels", None)
        out_channels = config.getoption("out_channels", None)
        num_train_timesteps = config.getoption("num_train_timesteps", 1000)
        num_infer_timesteps = config.getoption("num_infer_timesteps", 50)
        freeze_vae_encoder = config.getoption("freeze_vae_encoder", True)
        freeze_text_encoder = config.getoption("freeze_text_encoder", True)
        seed = config.getoption("seed", 1123)

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

        weight_path = config.getoption("pretrained_weight_path", None)

        if weight_path is None and pretrain_infos is not None:
            weight_path = [
                cached_path(nested_dict_value(pretrain_infos, "unet", "weight")),
                cached_path(nested_dict_value(pretrain_infos, "text", "weight")),
                cached_path(nested_dict_value(pretrain_infos, "vae", "weight")),
            ]

        inst.from_pretrained(weight_path)
        return inst

    @autocast()
    def forward(
        self,
    ):
        raise NotImplementedError

    @add_default_section_for_function("core/model/diffusers/inpainting/stable")
    @autocast()
    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        pixel_masks: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        strength: Optional[float] = 1.0,
        guidance_scale: Optional[float] = 7.5,
    ):
        outputs = super().generate(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            pixel_values=pixel_values,
            pixel_masks=pixel_masks,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
            strength=strength,
            guidance_scale=guidance_scale,
        )

        return DiffusionOutputs(outputs=outputs.images)


@register_model("core/model/diffusers/resolution/stable", diffusion_model_decorator)
class StableForImageResolution(_StableForImageResolution):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        quant_config_path: Optional[str] = None,
        image_size: Optional[int] = None,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        num_train_timesteps: Optional[int] = 1000,
        num_infer_timesteps: Optional[int] = 50,
        freeze_vae_encoder: Optional[bool] = True,
        freeze_text_encoder: Optional[bool] = True,
        seed: Optional[int] = 1123,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

    @classmethod
    @add_default_section_for_init("core/model/diffusers/resolution/stable")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/diffusers/resolution/stable")
        pretrained_name = config.getoption("pretrained_name", "stable-v1.5-x4-upscaler")
        pretrain_infos = nested_dict_value(pretrained_diffusers_infos, pretrained_name)

        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrain_infos, "unet", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = config.getoption("text_config_path", None)
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrain_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        vae_config_path = config.getoption("vae_config_path", None)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrain_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = config.getoption("scheduler_config_path", None)
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrain_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        quant_config_path = config.getoption("quant_config_path", None)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        image_size = config.getoption("image_size", None)
        in_channels = config.getoption("in_channels", None)
        out_channels = config.getoption("out_channels", None)
        num_train_timesteps = config.getoption("num_train_timesteps", 1000)
        num_infer_timesteps = config.getoption("num_infer_timesteps", 50)
        freeze_vae_encoder = config.getoption("freeze_vae_encoder", True)
        freeze_text_encoder = config.getoption("freeze_text_encoder", True)
        seed = config.getoption("seed", 1123)

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
            image_size=image_size,
            in_channels=in_channels,
            out_channels=out_channels,
            num_train_timesteps=num_train_timesteps,
            num_infer_timesteps=num_infer_timesteps,
            freeze_vae_encoder=freeze_vae_encoder,
            freeze_text_encoder=freeze_text_encoder,
            seed=seed,
        )

        weight_path = config.getoption("pretrained_weight_path", None)

        if weight_path is None and pretrain_infos is not None:
            weight_path = [
                cached_path(nested_dict_value(pretrain_infos, "unet", "weight")),
                cached_path(nested_dict_value(pretrain_infos, "text", "weight")),
                cached_path(nested_dict_value(pretrain_infos, "vae", "weight")),
            ]

        inst.from_pretrained(weight_path)
        return inst

    @autocast()
    def forward(
        self,
    ):
        raise NotImplementedError

    @add_default_section_for_function("core/model/diffusers/resolution/stable")
    @autocast()
    def generate(
        self,
        input_ids: torch.Tensor,
        negative_input_ids: torch.Tensor,
        pixel_values: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
        negative_attention_mask: Optional[torch.Tensor] = None,
        guidance_scale: Optional[float] = 9.0,
        noise_level: Optional[int] = 20,
    ):
        outputs = super().generate(
            input_ids=input_ids,
            negative_input_ids=negative_input_ids,
            pixel_values=pixel_values,
            attention_mask=attention_mask,
            negative_attention_mask=negative_attention_mask,
            guidance_scale=guidance_scale,
            noise_level=noise_level,
        )

        return DiffusionOutputs(outputs=outputs.images)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import re
import torch
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models.mistral import (
    MistralForGeneration as _MistralForGeneration,
)
from unitorch.models.mistral import MistralProcessor
from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models.mistral import pretrained_mistral_infos


class MistralForGenerationPipeline(_MistralForGeneration):
    def __init__(
        self,
        config_path: str,
        vocab_path: str,
        quant_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 512,
        max_gen_seq_length: Optional[int] = 512,
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        device: Optional[Union[str, int]] = "cpu",
    ):
        if device == "cpu":
            quant_config_path = None
        super().__init__(
            config_path=config_path,
            quant_config_path=quant_config_path,
        )
        self.processor = MistralProcessor(
            vocab_file=vocab_path,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self.to(device=self._device)
        self.eval()

    @classmethod
    @add_default_section_for_init("core/pipeline/mistral")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/pipeline/mistral")
        pretrained_name = config.getoption("pretrained_name", "default-mistral")

        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_mistral_infos, pretrained_name, "config"),
        )
        config_path = cached_path(config_path)

        vocab_path = config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_mistral_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        quant_config_path = config.getoption("quant_config_path", None)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        max_seq_length = config.getoption("max_seq_length", 512)
        max_gen_seq_length = config.getoption("max_gen_seq_length", 512)
        device = config.getoption("device", "cpu")
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_mistral_infos, pretrained_name, "weight"),
            check_none=False,
        )

        inst = cls(
            config_path,
            vocab_path=vocab_path,
            quant_config_path=quant_config_path,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
            weight_path=weight_path,
            device=device,
        )

        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/mistral")
    def __call__(
        self,
        prompt: str,
        max_seq_length: Optional[int] = 512,
        num_beams: Optional[int] = 2,
        decoder_start_token_id: Optional[int] = 1,
        decoder_end_token_id: Optional[Union[int, List[int]]] = [2],
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 512,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ):
        inputs = self.processor.prompt(
            text=prompt,
            max_seq_length=max_seq_length,
        )
        inputs = {k: v.unsqueeze(0) if v is not None else v for k, v in inputs.items()}
        inputs = {
            k: v.to(device=self._device) if v is not None else v
            for k, v in inputs.items()
        }
        outputs = super().generate(
            input_ids=inputs["input_ids"],
            num_beams=num_beams,
            decoder_start_token_id=decoder_start_token_id,
            decoder_end_token_id=decoder_end_token_id,
            num_return_sequences=num_return_sequences,
            min_gen_seq_length=min_gen_seq_length,
            max_gen_seq_length=max_gen_seq_length,
            repetition_penalty=repetition_penalty,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            do_sample=do_sample,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )
        decoded = self.processor.detokenize(outputs.sequences)

        return decoded[0].strip()
